'use strict';

var config = require('../../config.json');

var findComment = require('../lib/find-comment');
var checkFunction = require('../lib/check-function');
var parseJsDoc = require('../lib/parse-jsdoc');
var normalizeFunctionBody = require('../lib/normalize-function-body');
var insertParametersCheck = require('../lib/insert-parameters-check');
var returnStatementVisitorFactory = require('./return-statement');

var helpers = {
    isDeclarationIsFunction: function isDeclarationIsFunction(declaration) {
        return declaration.get('init').isFunction();
    },
    isPathIsConstructor: function isPathIsConstructor(path) {
        return path.node.kind === 'constructor';
    },
    isPathIsExpression: function isPathIsExpression(path) {
        return path.isExpressionStatement();
    },
    isPathIsClassDeclaration: function isPathIsClassDeclaration(parent) {
        return parent.isClassDeclaration();
    }
};

/**
 * @param {Function} typecheckFunctionCall
 * @param {Object} globalState
 * @param {Object} t
 * @returns {Object}
 */
module.exports = function (typecheckFunctionCall, globalState, t) {
    var returnTypecheckVisitor = returnStatementVisitorFactory(typecheckFunctionCall, globalState, t);

    /**
     * @param {String} comment
     * @param {NodePath} path
     * @param {String} [name]
     */
    function executeFunctionTransformation(comment, path, name) {
        var jsDoc = parseJsDoc(comment);

        if (!jsDoc) {
            return;
        }

        var functionName = name || jsDoc.name || (path.node.id ? path.node.id.name : config.defaultFunctionName);

        normalizeFunctionBody(path);

        if (!checkFunction(path)) {
            return;
        }

        insertParametersCheck(functionName, typecheckFunctionCall, path, jsDoc, globalState.useStrict, t);

        path.traverse(returnTypecheckVisitor, {
            jsDoc: jsDoc,
            functionPath: path,
            functionName: functionName
        });

        globalState.shouldInjectHelperFunction = true;
    }

    return {
        /**
         * @param {NodePath} path
         * @param {PluginPass} state
         */
        VariableDeclaration: function VariableDeclaration(path, state) {
            var comment = findComment(path, state, globalState.hasGlobalDirective);

            if (!comment) {
                return;
            }

            var declarations = path.get('declarations');
            var functionDeclaration = declarations.find(helpers.isDeclarationIsFunction);

            if (!functionDeclaration) {
                return;
            }

            executeFunctionTransformation(comment, functionDeclaration.get('init'));
        },


        /**
         * @param {NodePath} path
         * @param {PluginPass} state
         */
        ReturnStatement: function ReturnStatement(path, state) {
            var argument = path.get('argument');

            if (!argument || !argument.isFunction()) {
                return;
            }

            var comment = findComment(path, state, globalState.hasGlobalDirective);

            if (!comment) {
                return;
            }

            executeFunctionTransformation(comment, argument);
        },


        /**
         * @param {NodePath} path
         * @param {PluginPass} state
         */
        AssignmentExpression: function AssignmentExpression(path, state) {
            var rightPath = path.get('right');

            if (!rightPath.isFunction()) {
                return;
            }

            var expression = path.find(helpers.isPathIsExpression);
            var comment = findComment(expression, state, globalState.hasGlobalDirective);

            if (!comment) {
                return;
            }

            executeFunctionTransformation(comment, rightPath);
        },


        /**
         * @param {NodePath} path
         * @param {PluginPass} state
         */
        ObjectProperty: function ObjectProperty(path, state) {
            var value = path.get('value');

            if (!value.isFunction()) {
                return;
            }

            var comment = findComment(path, state, globalState.hasGlobalDirective);

            if (!comment) {
                return;
            }

            executeFunctionTransformation(comment, value, path.node.key.name);
        },


        /**
         * @param {NodePath} path
         * @param {PluginPass} state
         */
        ClassMethod: function ClassMethod(path, state) {
            var comment = findComment(path, state, globalState.hasGlobalDirective);

            if (!comment) {
                return;
            }

            var classDeclaration = path.find(helpers.isPathIsClassDeclaration);
            var className = classDeclaration.node.id.name;
            var node = path.node;
            var kind = node.kind === 'constructor' ? '' : node.kind + ' ';

            executeFunctionTransformation(comment, path, '' + (node.static ? 'static ' : '') + kind + className + '.' + node.key.name);
        },


        /**
         * @param {NodePath} path
         * @param {PluginPass} state
         */
        ObjectMethod: function ObjectMethod(path, state) {
            var comment = findComment(path, state, globalState.hasGlobalDirective);

            if (!comment) {
                return;
            }

            executeFunctionTransformation(comment, path, path.node.key.name);
        },


        /**
         * @param {NodePath} path
         * @param {PluginPass} state
         */
        'FunctionDeclaration|ArrowFunctionExpression': function FunctionDeclarationArrowFunctionExpression(path, state) {
            var comment = findComment(path, state, globalState.hasGlobalDirective);

            if (!comment) {
                return;
            }

            executeFunctionTransformation(comment, path);
        }
    };
};